"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""
import logging
import platform
import tempfile

import distro

from pysolbase.SolBase import SolBase

logger = logging.getLogger(__name__)


class PlatformTools(object):
    """
    Platform tools
    """

    @classmethod
    def is_os_64(cls):
        """
        Return true is platform is 64 bits
        :return bool
        :rtype bool
        """

        try:
            return platform.machine().find('64') >= 0
        except Exception as e:
            # Not normal, but fallback 64 bits
            logger.warning("Ex=%s", SolBase.extostr(e))
            return True

    @classmethod
    def is_cpu_arm(cls):
        """
        Return true is cpu is an arm (32 or 64 bits)
        :return bool
        :rtype bool
        """

        try:
            if platform.machine().lower().startswith("arm") or platform.machine().lower().startswith("aarch"):
                return True
            else:
                return False
        except Exception as e:
            # Not normal, but fallback x86
            logger.warning("Ex=%s", SolBase.extostr(e))
            return False

    @classmethod
    def get_distribution_type(cls):
        """
        Debian      : return debian
        Ubuntu      : return debian
        Centos      : return redhat
        Redhat      : return redhat
        Raspbian    : return debian
        Windows     : return windows
        Unknown     : return debian + warn log

        *** Ubuntu 15
        platform.linux_distribution(full_distribution_name=False)
        ('Ubuntu', '15.10', 'wily')
        platform.dist()
        ('Ubuntu', '15.10', 'wily')

        *** Debian 8
        platform.linux_distribution(full_distribution_name=False)
        ('Debian', '8', '')
        platform.dist()
        ('Debian', '8', '')

        *** Centos 7
        platform.linux_distribution(full_distribution_name=False)
        ('centos', '7.3.1611', 'Core')
        platform.dist()
        ('centos', '7.3.1611', 'Core')

        *** Raspbian 8
        platform.linux_distribution(full_distribution_name=False)
        ('Debian', '8', '')
        platform.dist()
        ('Debian', '8', '')

        *** Usefull : https://github.com/hpcugent/easybuild/wiki/OS_flavor_name_version

        :return: str
        :rtype str
        """

        # Detected
        detected_dist = None

        # Get
        cur_linux_distribution = distro.linux_distribution(full_distribution_name=False)

        # Try linux_distribution if required
        if not detected_dist:
            try:
                detected_dist = cur_linux_distribution[0]
                detected_dist = detected_dist.strip()
                if len(detected_dist) == 0:
                    # Reset
                    detected_dist = None
            except Exception as e:
                logger.debug("Ex=%s", SolBase.extostr(e))

        # This is a fallback windows
        if not detected_dist:
            # Windows HACK
            detected_dist = platform.system()

        # Ok
        return cls._get_distribution_type(detected_dist,  cur_linux_distribution)

    @classmethod
    def _get_distribution_type(cls, detected_dist,  cur_linux_distribution):
        """
        Internal method
        :param detected_dist: str,None
        :type detected_dist: str,None
        :param cur_linux_distribution: list,tuple, for log only
        :type cur_linux_distribution: list,tuple
        :return str
        :rtype str
        """

        # Check
        if not detected_dist:
            logger.warning("Unable to detect distribution, fallback debian, got detected_dist=%s, cur_linux_distribution=%s", detected_dist, cur_linux_distribution)
            return "debian"

        # Lower
        detected_dist = detected_dist.lower()

        # Check
        if detected_dist == "debian":
            return "debian"
        elif detected_dist == "ubuntu":
            return "debian"
        elif detected_dist == "centos":
            return "redhat"
        elif detected_dist == "scientific linux":
            return "redhat"

        # Here some hacks (zzz)
        elif detected_dist.find("redhat") >= 0:
            return "redhat"
        elif detected_dist.find("red hat") >= 0:
            return "redhat"
        elif detected_dist.find("centos") >= 0:
            return "redhat"
        elif detected_dist.find("rehl") >= 0:
            return "redhat"

        # Windows
        elif detected_dist.find("windows") >= 0:
            return "windows"

        # Fallback
        logger.warning("Unknown distribution, fallback debian, got detected_dist=%s, cur_linux_distribution=%s", detected_dist, cur_linux_distribution)
        return "debian"

    @classmethod
    def get_tmp_dir(cls):
        """
        Get tmp dir
        """
        if cls.get_distribution_type() == "windows":
            return tempfile.gettempdir()
        else:
            return "/tmp"



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""


class Assert(object):
    """
    Assert helpers
    """

    @classmethod
    def check(cls, exception_class, condition, message, *args, **kwargs):
        """
        Check a condition and, if false, raise provided exception with provided message.
        :param exception_class: Exception class to raise if condition is False.
        :type exception_class: callable
        :param condition: Condition to evaluate. If false, will raise. If dict|tuple|list is empty, will raise.
        :type condition: bool,list,tuple,dict
        :param message: Exception message
        :type message: basestring
        """

        if condition:
            return

        raise exception_class(message, *args, **kwargs)



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""
from logging import Filter

from gevent.local import local

from pysolbase.Assert import Assert


class ContextFilter(Filter):
    """
    Context filter
    DOC : https://docs.python.org/3/howto/logging-cookbook.html#adding-contextual-information-to-your-logging-output
    """

    # CAUTION : We must target local here as global (otherwise its a variable local to functions)
    LOC = local()

    @classmethod
    def set_value(cls, k, v):
        """
        Set context value
        :param k: key name
        :type k: basestring
        :param v: value
        :type v: object
        """

        Assert.check(Exception, k is not None, "Need k, got None")
        Assert.check(Exception, len(k) > 0, "Need k, got empty")

        setattr(cls.LOC, k, v)

    def filter(self, record):
        """
        Record filter.
        This will push thread context (using LOC) toward logger item "kfilter", as an OrderedDict, formatted as "key0:value0 keyN:valueN"
        :param record: logging.LogRecord
        :type record: logging.LogRecord
        :return: bool
        :rtype bool
        """

        # Prepare
        s = u""
        for k in sorted(self.LOC.__dict__.keys()):
            v = self.LOC.__dict__[k]
            s += u" %s:%s" % (k, v)
        s += u" "

        # Push to record in a single shot
        setattr(record, "kfilter", s)

        return True



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""
import ast
import logging
import os
import platform
import sys
import time
import traceback
from logging.config import dictConfig
from logging.handlers import WatchedFileHandler, TimedRotatingFileHandler, SysLogHandler
from threading import Lock

import gevent
import pytz
from datetime import datetime

from gevent import monkey, config
from yaml import load, SafeLoader

from pysolbase.ContextFilter import ContextFilter

logger = logging.getLogger(__name__)
lifecyclelogger = logging.getLogger("lifecycle")


class SolBase(object):
    """
    Base utilities & helpers.
    """

    # ===============================
    # STATIC STUFF
    # ===============================

    # Component name (mainly for rsyslog)
    _compo_name = "CompoNotSet"

    # Global init stuff
    _voodoo_initialized = False
    _voodoo_lock = Lock()

    # Logging stuff
    _logging_initialized = False
    _logging_lock = Lock()

    # Fork stuff
    _master_process = True

    # ===============================
    # DATE & MS
    # ===============================

    @classmethod
    def mscurrent(cls):
        """
        Return current millis since epoch
        :return float
        :rtype float
        """
        return time.time() * 1000.0

    @classmethod
    def securrent(cls):
        """
        Return current seconds since epoch
        :return float
        :rtype float
        """
        return time.time()

    @classmethod
    def msdiff(cls, ms_start, ms_end=None):
        """
        Get difference in millis between current millis and provided millis.
        :param ms_start: Start millis
        :type ms_start: float
        :param ms_end: End millis (will use current if not provided)
        :type ms_end: float
        :return float
        :rtype float
        """

        if not ms_end:
            ms_end = cls.mscurrent()

        return ms_end - ms_start

    @classmethod
    def datecurrent(cls, erase_mode=0):
        """
        Return current date (UTC)
        :param erase_mode: Erase mode (0=nothing, 1=remove microseconds but keep millis, 2=remove millis completely)
        :return datetime.datetime
        :rtype datetime.datetime
        """

        if erase_mode == 0:
            return datetime.utcnow()
        elif erase_mode == 1:
            # Force precision loss (keep millis, kick micro)
            dt = datetime.utcnow()
            return dt.replace(microsecond=int((dt.microsecond * 0.001) * 1000))
        elif erase_mode == 2:
            return datetime.utcnow().replace(microsecond=0)

    @classmethod
    def datediff(cls, dt_start, dt_end=None):
        """
        Get difference in millis between two datetime
        :param dt_start: Start datetime
        :type dt_start: datetime.datetime
        :param dt_end: End datetime (will use current utc if not provided)
        :type dt_end: datetime.datetime
        :return float
        :rtype float
        """

        # Fix
        if not dt_end:
            dt_end = cls.datecurrent()

        # Get delta
        delta = dt_end - dt_start
        return ((delta.days * 86400 + delta.seconds) * 1000) + (delta.microseconds * 0.001)

    # ==========================================
    # EPOCH / DT
    # ==========================================

    DT_EPOCH = datetime.utcfromtimestamp(0)

    @classmethod
    def dt_to_epoch(cls, dt):
        """
        Convert a datetime (UTC required) to a unix time since epoch, as seconds, as integer.
        Note that millis precision is lost.
        :param dt: datetime
        :type dt: datetime
        :return int
        :rtype int
        """

        return int((dt - cls.DT_EPOCH).total_seconds())

    @classmethod
    def epoch_to_dt(cls, epoch):
        """
        Convert an epoch float or int to datetime (UTC)
        :param epoch: float,int
        :type epoch: float,int
        :return datetime
        :rtype datetime
        """

        return datetime.utcfromtimestamp(epoch)

    @classmethod
    def dt_is_naive(cls, dt):
        """
        Return true if dt is naive
        :param dt: datetime.datetime
        :type dt: datetime.datetime
        :return bool
        :rtype bool
        """

        # Naive : no tzinfo
        if not dt.tzinfo:
            return True

        # Aware
        return False

    @classmethod
    def dt_ensure_utc_aware(cls, dt):
        """
        Switch dt to utc time zone. If dt is naive, assume utc, otherwise, convert it to utc timezone.
        Return an AWARE timezone (utc switched) datetime,
        :param dt: datetime.datetime
        :type dt:datetime.datetime
        :return datetime.datetime
        :rtype datetime.datetime
        """

        # If naive, add utc
        if cls.dt_is_naive(dt):
            return dt.replace(tzinfo=pytz.utc)
        else:
            # Not naive, go utc, keep aware
            return dt.astimezone(pytz.utc)

    @classmethod
    def dt_ensure_utc_naive(cls, dt):
        """
        Ensure dt is naive. Return dt, switched to UTC (if applicable), and naive.
        :param dt: datetime.datetime
        :type dt:datetime.datetime
        :return datetime.datetime
        :rtype datetime.datetime
        """

        dt = cls.dt_ensure_utc_aware(dt)
        return dt.replace(tzinfo=None)

    # ===============================
    # COMPO NAME (FOR RSYSLOG)
    # ===============================

    @classmethod
    def set_compo_name(cls, compo_name):
        """
        Set the component name. Useful for rsyslog.
        :param compo_name: The component name or None. If None, method do nothing.
        :type compo_name: str,None
        """

        if compo_name:
            cls._compo_name = compo_name
            lifecyclelogger.debug("compo_name now set to=%s", cls._compo_name)

    @classmethod
    def get_compo_name(cls):
        """
        Get current component name.
        :return str
        :rtype str
        """

        return cls._compo_name

    @classmethod
    def get_machine_name(cls):
        """
        Get machine name
        :return: Machine name
        :rtype: str
        """

        return platform.uname()[1]

    # ===============================
    # MISC
    # ===============================

    @classmethod
    def sleep(cls, sleep_ms):
        """
        Sleep for specified ms.
        Also used as gevent context switch in code, since it rely on gevent.sleep.
        :param sleep_ms: Millis to sleep.
        :type sleep_ms: int,float
        :return Nothing.
        """
        ms = sleep_ms * 0.001

        # gevent 1.3 : ms is not fully respected (100 can be 80-100)
        gevent.sleep(ms)

    # ===============================
    # EXCEPTION HELPER
    # ===============================

    @classmethod
    def extostr(cls, e, max_level=30, max_path_level=5):
        """
        Format an exception.
        :param e: Any exception instance.
        :type e: BaseException
        :param max_level: Maximum call stack level (default 30)
        :type max_level: int
        :param max_path_level: Maximum path level (default 5)
        :type max_path_level: int
        :return The exception readable string
        :rtype str
        """

        # Go
        list_frame = None
        try:
            out_buffer = ""

            # Class type
            out_buffer += "e.cls:[{0}]".format(e.__class__.__name__)

            # To string
            try:
                ex_buf = str(e)
            except UnicodeEncodeError:
                ex_buf = repr(str(e))
            except Exception as e:
                logger.warning("Exception, e=%s", e)
                raise
            out_buffer += ", e.bytes:[{0}]".format(ex_buf)

            # Traceback
            si = sys.exc_info()

            # Raw frame
            # tuple : (file, lineno, method, code)
            raw_frame = traceback.extract_tb(si[2])
            raw_frame.reverse()

            # Go to last tb_next
            last_tb_next = None
            cur_tb = si[2]
            while cur_tb:
                last_tb_next = cur_tb
                cur_tb = cur_tb.tb_next

            # Skip frame up to current raw frame count
            list_frame = list()
            cur_count = -1
            skip_count = len(raw_frame)
            if last_tb_next:
                cur_frame = last_tb_next.tb_frame
            else:
                cur_frame = None
            while cur_frame:
                cur_count += 1
                if cur_count < skip_count:
                    cur_frame = cur_frame.f_back
                else:
                    # Need : tuple : (file, lineno, method, code)
                    raw_frame.append((cur_frame.f_code.co_filename, cur_frame.f_lineno, cur_frame.f_code.co_name, ""))
                    cur_frame = cur_frame.f_back

            # Build it
            cur_idx = 0
            out_buffer += ", e.cs=["
            for tu in raw_frame:
                line = tu[1]
                cur_file = tu[0]
                method = tu[2]

                # Handle max path level
                ar_token = cur_file.rsplit(os.sep, max_path_level)
                if len(ar_token) > max_path_level:
                    # Remove head
                    ar_token.pop(0)
                    # Join
                    cur_file = "..." + os.sep.join(ar_token)

                # Format
                out_buffer += "in:{0}#{1}@{2} ".format(method, cur_file, line)

                # Loop
                cur_idx += 1
                if cur_idx >= max_level:
                    out_buffer += "..."
                    break

            # Close
            out_buffer += "]"

            # Ok
            return out_buffer
        finally:
            if list_frame:
                del list_frame

    # ===============================
    # VOODOO INIT
    # ===============================

    @classmethod
    def _reset(cls):
        """
        For unittest only
        """

        cls._logging_initialized = False
        cls._voodoo_initialized = False

    @classmethod
    def voodoo_init(cls, aggressive=True, init_logging=True):
        """
        Global initialization, to call asap.
        Apply gevent stuff & default logging configuration.
        :param aggressive: bool
        :type aggressive: bool
        :param init_logging: If True, logging_init is called.
        :type init_logging: bool
        :return Nothing.
        """

        try:
            # Check
            if cls._voodoo_initialized:
                return

            # Lock
            with cls._voodoo_lock:
                # Re-check
                if cls._voodoo_initialized:
                    return

                # Fire the voodoo magic :)
                lifecyclelogger.debug("Voodoo : gevent : entering, aggressive=%s", aggressive)
                monkey.patch_all(aggressive=aggressive)
                lifecyclelogger.debug("Voodoo : gevent : entering")

                # Gevent 1.3 : by default, gevent keep tracks of spawn call stack
                # This may lead to memory leak, if a method spawn itself in loop (timer mode)
                # We disable this
                config.track_greenlet_tree = False

                # Initialize log level to INFO
                if init_logging:
                    lifecyclelogger.debug("Voodoo : logging : entering")
                    cls.logging_init()
                    lifecyclelogger.debug("Voodoo : logging : done")

                # Done
                cls._voodoo_initialized = True
        finally:
            # If whenever init_logging if set AND it is NOT initialized => we must init it
            # => we may have been called previously with init_logging=false, but monkey patch is SET and logging not initialized
            # => so it must be init now
            if init_logging and not cls._logging_initialized:
                lifecyclelogger.debug("Voodoo : logging : not yet init : entering")
                cls.logging_init()
                lifecyclelogger.debug("Voodoo : logging : not yet init : done")

    # ===============================
    # LOGGING
    # ===============================

    @classmethod
    def logging_init(cls, log_level="INFO", force_reset=False, log_callback=None,
                     log_to_file=None,
                     log_to_syslog=True,
                     log_to_syslog_facility=SysLogHandler.LOG_LOCAL0,
                     log_to_console=True,
                     log_to_file_mode="watched_file",
                     context_filter=None):
        """
        Initialize logging sub system with default settings (console, pre-formatted output)
        :param log_to_console: if True to console
        :type log_to_console: bool
        :param log_level: The log level to set. Any value in "DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"
        :type log_level: str
        :param force_reset: If true, logging system is reset.
        :type force_reset: bool
        :param log_to_file: If specified, log to file
        :type log_to_file: str,None
        :param log_to_syslog: If specified, log to syslog
        :type log_to_syslog: bool
        :param log_to_syslog_facility: Syslog facility.
        :type log_to_syslog_facility: int
        :param log_to_file_mode: str "watched_file" for WatchedFileHandler, "time_file" for TimedRotatingFileHandler (or time_file_seconds for unittest)
        :type log_to_file_mode: str
        :param log_callback: Callback for unittest
        :param context_filter: Context filter. If None, pysolbase.ContextFilter.ContextFilter is used. If used instance has an attr "filter", it is added to all handlers and "%(kfilter)s" will be populated by all thread context key/values, using filter method call. Refer to our ContextFilter default implementation for details.
        :type context_filter: None,object
        :return Nothing.
        """

        if cls._logging_initialized and not force_reset:
            return

        with cls._logging_lock:
            if cls._logging_initialized and not force_reset:
                return

            # Reset
            cls._reset_logging(log_level=log_level)

            # Default
            logging.basicConfig(level=log_level)

            # Filter
            if context_filter:
                c_filter = context_filter
            else:
                c_filter = ContextFilter()

            # Format begin
            s_f = "%(asctime)s | %(levelname)s | %(module)s@%(funcName)s@%(lineno)d | %(message)s "

            # Browse
            if hasattr(c_filter, "filter"):
                # Push generic field
                # We expect it to be formatted like our pysolbase.ContextFilter.ContextFilter#filter method.
                s_f += "|%(kfilter)s"

            # Format end
            s_f += "| %(thread)d:%(threadName)s | %(process)d:%(processName)s"

            # Formatter
            f = logging.Formatter(s_f)

            # Console handler
            c = None
            if log_to_console:
                # This can be overriden by unittest, we use __stdout__
                c = logging.StreamHandler(sys.__stdout__)
                c.setLevel(logging.getLevelName(log_level))
                c.setFormatter(f)

            # File handler to /tmp
            cf = None
            if log_to_file:
                if log_to_file_mode == "watched_file":
                    cf = WatchedFileHandler(log_to_file, encoding="utf-8")
                    cf.setLevel(logging.getLevelName(log_level))
                    cf.setFormatter(f)
                elif log_to_file_mode == "time_file":
                    cf = TimedRotatingFileHandler(log_to_file, encoding="utf-8", utc=True, when="D", interval=1, backupCount=7)
                    cf.setLevel(logging.getLevelName(log_level))
                    cf.setFormatter(f)
                elif log_to_file_mode == "time_file_seconds":
                    # For unittest only
                    cf = TimedRotatingFileHandler(log_to_file, encoding="utf-8", utc=True, when="S", interval=1, backupCount=7)
                    cf.setLevel(logging.getLevelName(log_level))
                    cf.setFormatter(f)
                else:
                    logger.warning("Invalid log_to_file_mode=%s", log_to_file_mode)

            # Syslog handler
            syslog = None
            if log_to_syslog:
                try:
                    from pysolbase.SysLogger import SysLogger

                    syslog = SysLogger(log_callback=log_callback, facility=log_to_syslog_facility)
                    syslog.setLevel(logging.getLevelName(log_level))
                    syslog.setFormatter(f)
                except Exception as e:
                    # This will fail on WINDOWS (no attr AF_UNIX)
                    logger.debug("Unable to import SysLogger, e=%s", SolBase.extostr(e))
                    syslog = False

            # Initialize
            root = logging.getLogger()
            root.setLevel(logging.getLevelName(log_level))
            root.handlers = []
            if log_to_console:
                c.addFilter(c_filter)
                root.addHandler(c)
            if log_to_file and cf:
                cf.addFilter(c_filter)
                root.addHandler(cf)
            if log_to_syslog and syslog:
                syslog.addFilter(c_filter)
                root.addHandler(syslog)

            # Done
            cls._logging_initialized = True
            if force_reset:
                lifecyclelogger.info("Logging : initialized from memory, log_level=%s, force_reset=%s", log_level, force_reset)
            else:
                lifecyclelogger.debug("Logging : initialized from memory, log_level=%s, force_reset=%s", log_level, force_reset)

    @classmethod
    def _register_filter(cls, c_filter):
        """
        Register filter across the whole logging (root and all loggers)
        Notice : addFilter is protected against duplicates add
        :param c_filter: pysolbase.ContextFilter.ContextFilter
        :type c_filter: pysolbase.ContextFilter.ContextFilter
        """

        # Initialize
        root = logging.getLogger()
        for h in list(root.handlers):
            h.addFilter(c_filter)

        # Browse all loggers and set
        for name in logging.root.manager.loggerDict:
            cur_logger = logging.getLogger(name)
            for h in list(cur_logger.handlers):
                h.addFilter(c_filter)

    @classmethod
    def _reset_logging(cls, log_level):
        """
        Reset
        :param log_level: str
        :type log_level: str
        """

        # Found no way to fully reset the logging stuff while running
        # We reset root and all loggers to INFO, and kick handlers

        # Initialize
        root = logging.getLogger()
        root.setLevel(logging.getLevelName(log_level))
        for h in root.handlers:
            # noinspection PyBroadException
            try:
                h.close()
            except:
                pass
        root.handlers = []

        # Browse all loggers and set
        for name in logging.root.manager.loggerDict:
            cur_logger = logging.getLogger(name)
            cur_logger.setLevel(logging.getLevelName(log_level))
            for h in cur_logger.handlers:
                # noinspection PyBroadException
                try:
                    h.close()
                except:
                    pass
            cur_logger.handlers = []

    @classmethod
    def logging_initfromfile(cls, config_file_name, force_reset=False, context_filter=None):
        """
        Initialize logging system from a configuration file, with optional reset.
        :param config_file_name: Configuration file name
        :type config_file_name: str
        :param force_reset: If true, logging system is reset.
        :type force_reset: bool
        :param context_filter: Context filter. If None, pysolbase.ContextFilter.ContextFilter is used. If used instance has an attr "filter", it is added to all handlers and "%(kfilter)s" will be populated by all thread context key/values, using filter method call. Refer to our ContextFilter default implementation for details.
        :type context_filter: None,object
        :return Nothing.
        """

        if cls._logging_initialized and not force_reset:
            return

        with cls._logging_lock:
            if cls._logging_initialized and not force_reset:
                return

            try:
                # Filter
                if context_filter:
                    c_filter = context_filter
                else:
                    c_filter = ContextFilter()

                # Reset
                cls._reset_logging(log_level="INFO")

                # Load
                logger.debug("Logging : yaml config_file_name=%s", config_file_name)
                with open(config_file_name, 'r') as f:
                    d = load(f, Loader=SafeLoader)
                    dictConfig(d)

                # Register filter
                if c_filter:
                    cls._register_filter(c_filter)

                if force_reset:
                    lifecyclelogger.info("Logging : initialized from yaml file, config_file_name=%s", config_file_name)
                else:
                    lifecyclelogger.debug("Logging : initialized from yaml file, config_file_name=%s", config_file_name)
            except Exception:
                raise

    @classmethod
    def context_set(cls, k, v):
        """
        Set thread/greenlet context value

        This is a wrapper to pysolbase.ContextFilter.ContextFilter#set_value
        and will work only if ContextFilter is defined (which is by default)
        :param k: key name
        :type k: basestring
        :param v: value
        :type v: object
        """

        ContextFilter.set_value(k, v)

    # ===============================
    # FORK STUFF
    # ===============================

    @classmethod
    def get_master_process(cls):
        """
        Return True if we are the master process, False otherwise.
        :return bool
        :rtype bool
        """
        return cls._master_process

    @classmethod
    def set_master_process(cls, b):
        """
        Set is we are a fork master or not
        :param b: True if we are master process, False if we are a child process.
        :type b: bool
        :return Nothing
        """

        logger.debug("Switching _masterProcess to %s", b)
        cls._master_process = b

    # ===============================
    # BINARY STUFF
    # ===============================

    @classmethod
    def binary_to_unicode(cls, bin_buf, encoding="utf-8"):
        """
        Binary buffer to str, using the specified encoding
        :param bin_buf: Binary buffer
        :type bin_buf: bytes
        :param encoding: Encoding to use
        :type encoding: str
        :return str
        :rtype str
        """

        return bin_buf.decode(encoding)

    @classmethod
    def unicode_to_binary(cls, unicode_buf, encoding="utf-8"):
        """
        Unicode to binary buffer, using the specified encoding
        :param unicode_buf: String to convert.
        :type unicode_buf: str
        :param encoding: Encoding to use.
        :type encoding: str
        :return bytes
        :rtype bytes
        """

        return unicode_buf.encode(encoding)

    @classmethod
    def fix_paths_for_popen(cls):
        """
        Fix path and env for popen calls toward current project
        Mainly used for unittests, which requires current env to be propagated while testing command line invocation within same project
        """

        # Merge all
        ar_p_path = sys.path
        if os.environ.get("PYTHONPATH"):
            ar_p_path.extend(os.environ.get("PYTHONPATH").split(":"))
        if os.environ.get("PATH"):
            ar_p_path.extend(os.environ.get("PATH").split(":"))

        # Join
        new_path = ":".join(ar_p_path)

        # Re-Assign
        os.environ["PATH"] = new_path
        os.environ["PYTHONPATH"] = new_path

    # ===============================
    # CONVERSIONS
    # ===============================

    @classmethod
    def to_int(cls, v):
        """
        Convert to int
        :param v: int,str
        :type v: int,str
        :return: int
        :rtype int
        """

        if isinstance(v, int):
            return v
        else:
            return int(v)

    @classmethod
    def to_bool(cls, v):
        """
        Convert to bool
        :param v: bool,str
        :type v: bool,str
        :return: bool
        :rtype bool
        """

        if isinstance(v, bool):
            return v
        else:
            return ast.literal_eval(v)

    @classmethod
    def get_classname(cls, my_instance):
        """
        Return the class name of my_instance, or "Instance.None".
        :param cls: Our class.
        :param my_instance: Instance to use.
        :return: Return the class name of my_instance, or "Instance.None" in case of error/None value.
        """
        if my_instance is None:
            return "Instance.None"
        else:
            return my_instance.__class__.__name__

    @classmethod
    def get_pathseparator(cls):
        """
        Return the path separator.
        https://docs.python.org/library/os.html#os.sep
        :param cls: Our class
        :return: The path separator (string)
        """
        return os.sep

    @classmethod
    def is_bool(cls, my_bool):
        """
        Return true if the provided my_bool is a boolean.
        :param cls: Our class.
        :param my_bool: A boolean..
        :return: Return true if the provided my_bool is a boolean. False otherwise.
        """
        if my_bool is None:
            return False
        else:
            return isinstance(my_bool, bool)

    @classmethod
    def is_int(cls, my_int):
        """
        Return true if the provided my_int is a integer.
        :param cls: Our class.
        :param my_int: An integer..
        :return: Return true if the provided my_int is a integer. False otherwise.
        """
        if my_int is None:
            return False
        # Caution, boolean is an integer...
        elif SolBase.is_bool(my_int):
            return False
        else:
            return isinstance(my_int, int)

    @classmethod
    def get_current_pid_as_string(cls):
        """
        Return the current pids as string.
        :param cls: Our class.
        :return: A String
        """
        try:
            return "pid={0}, ppid={1}".format(os.getpid(), os.getppid())
        except AttributeError:
            return "pid={0}".format(os.getpid())

    # =====================================================
    # HELPER FOR SOCKET CLOSING
    # =====================================================

    @classmethod
    def safe_close_socket(cls, soc_to_close):
        """
        Safe close a socket
        :param soc_to_close: socket
        :type soc_to_close: socket.socket
        """

        if soc_to_close is None:
            return

        try:
            soc_to_close.shutdown(2)
        except Exception as e:
            logger.debug("Socket shutdown ex=%s", SolBase.extostr(e))

        try:
            soc_to_close.close()
        except Exception as e:
            logger.debug("Socket close ex=%s", SolBase.extostr(e))

        try:
            del soc_to_close
        except Exception as e:
            logger.debug("Socket del ex=%s", SolBase.extostr(e))



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""


class FlagTools(object):
    """
    Flag tools
    """

    @classmethod
    def is_set(cls, val, flags):
        """
        Check if all bit in flags are set in val
        :param val: int
        :type val: int
        :param flags: int
        :type flags: int
        :return bool
        :rtype bool
        """

        if (val & flags) == flags:
            return True
        else:
            return False

    @classmethod
    def set(cls, val, flags):
        """
        Add all bits in flags inside val, return new val
        :param val: int
        :type val: int
        :param flags: int
        :type flags: int
        :return int
        :rtype int
        """

        return val | flags

    @classmethod
    def clear(cls, val, flags):
        """
        Clear all bits in flags inside val, return new val
        :param val: int
        :type val: int
        :param flags: int
        :type flags: int
        :return int
        :rtype int
        """
        return val & ~flags



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

# Import
import logging

import os
import codecs

from pysolbase.SolBase import SolBase

logger = logging.getLogger(__name__)


class FileUtility(object):
    """
    File utility
    """

    @staticmethod
    def is_path_exist(path_name):
        """
        Check if a path (file or dir) name exist.
        :param path_name: Path name.
        :type path_name text_type
        :return: Return true (exist), false (do not exist, or invalid file name)
        :rtype bool
        """

        # Check
        if path_name is None:
            logger.warning("is_path_exist : file_name is None")
            return False
        elif not isinstance(path_name, str):
            logger.warning("is_path_exist : path_name not a text_type, className=%s", SolBase.get_classname(path_name))
            return False

        # Go
        return os.path.exists(path_name)

    @staticmethod
    def is_file_exist(file_name):
        """
        Check if file name exist.
        :param file_name: File name.
        :type file_name: str
        :return: Return true (exist), false (do not exist, or invalid file name)
        :rtype bool
        """

        # Check
        if file_name is None:
            logger.warning("is_file_exist : file_name is None")
            return False
        elif not isinstance(file_name, str):
            logger.warning("is_file_exist : file_name not a text_type, className=%s", SolBase.get_classname(file_name))
            return False

        # Go
        return os.path.isfile(file_name)

    @staticmethod
    def is_dir_exist(dir_name):
        """
        Check if dir name exist.
        :param dir_name: Directory name.
        :type dir_name: str
        :return: Return true (exist), false (do not exist, or invalid file name)
        :rtype bool
        """

        # Check
        if dir_name is None:
            logger.warning("is_dir_exist : file_name is None")
            return False
        elif not isinstance(dir_name, str):
            logger.warning("is_dir_exist : file_name not a text_type, className=%s", SolBase.get_classname(dir_name))
            return False

        # Go
        return os.path.isdir(dir_name)

    @staticmethod
    def get_file_size(file_name):
        """
        Return a file size in bytes.
        :param file_name: File name.
        :type file_name: str
        :return: An integer, gt-eq 0 if file exist, lt 0 if error.
        :rtype int
        """
        if not FileUtility.is_file_exist(file_name):
            return -1
        else:
            return os.path.getsize(file_name)

    @classmethod
    def get_current_dir(cls):
        """
        Return the current directory.
        :return: A String
        :rtype text_type
        """

        return os.getcwd()

    @staticmethod
    def file_to_binary(file_name):
        """
        Load a file toward a binary buffer.
        :param file_name: File name.
        :type file_name: str
        :return: Return the binary buffer or None in case of error.
        :rtype: bytes,None
        """

        # Check
        if not FileUtility.is_file_exist(file_name):
            logger.warning("file_to_binary : file_name not exist, file_name=%s", file_name)
            return None

        # Go
        rd = None
        try:
            # Open (binary : open return a io.BufferedReader)
            rd = open(file_name, "rb")

            # Read everything
            return rd.read()
        except IOError as e:
            # Exception...
            logger.warning("IOError, ex=%s", SolBase.extostr(e))
            return None
        except Exception as e:
            logger.warning("Exception, ex=%s", SolBase.extostr(e))
            return None
        finally:
            # Close if not None...
            if rd:
                rd.close()

    @staticmethod
    def file_to_textbuffer(file_name, encoding):
        """
        Load a file toward a text buffer (UTF-8), using the specify encoding while reading.
        CAUTION : This will read the whole file IN MEMORY.
        :param file_name: File name.
        :type file_name: str
        :param encoding: Encoding to use.
        :type encoding: str
        :return: A text buffer or None in case of error.
        :rtype str
        """

        # Check
        if not FileUtility.is_file_exist(file_name):
            logger.warning("file_to_textbuffer : file_name not exist, file_name=%s", file_name)
            return None

        # Go
        rd = None
        try:
            # Open (text : open return a io.BufferedReader)
            rd = codecs.open(file_name, "r", encoding, "strict", -1)

            # Read everything
            return rd.read()
        except IOError as e:
            # Exception...
            logger.warning("file_to_binary : IOError, ex=%s", SolBase.extostr(e))
            return None
        except Exception as e:
            logger.warning("file_to_binary : Exception, ex=%s", SolBase.extostr(e))
            return None
        finally:
            # Close if not None...
            if rd:
                rd.close()

    @staticmethod
    def append_binary_to_file(file_name, bin_buf):
        """
        Write to the specified filename, the provided binary buffer.
        Create the file if required.
        :param file_name:  File name.
        :type file_name: str
        :param bin_buf: Binary buffer to write.
        :type bin_buf: bytes
        :return: The number of bytes written or lt 0 if error.
        :rtype int
        """

        # Go
        rd = None
        try:
            # Open (text : open return a io.BufferedReader)
            rd = open(file_name, "ab+")

            # Read everything
            return rd.write(bin_buf)
        except IOError as e:
            # Exception...
            logger.warning("append_binary_to_file : IOError, ex=%s", SolBase.extostr(e))
            return -1
        except Exception as e:
            logger.warning("append_binary_to_file : Exception, ex=%s", SolBase.extostr(e))
            return -1
        finally:
            # Close if not None...
            if rd:
                rd.close()

    @staticmethod
    def append_text_to_file(file_name, text_buffer, encoding, overwrite=False):
        """
        Write to the specified filename, the provided binary buffer
        Create the file if required.
        :param file_name:  File name.
        :type file_name: str
        :param text_buffer: Text buffer to write.
        :type text_buffer: str
        :param encoding: The encoding to use.
        :type encoding: str
        :param overwrite: If true, file is overwritten.
        :type overwrite: bool
        :return: The number of bytes written or lt 0 if error.
        :rtype int
        """

        # Go
        rd = None
        try:
            # Open (text : open return a io.BufferedReader)
            if not overwrite:
                rd = codecs.open(file_name, "a+", encoding, "strict", -1)
            else:
                rd = codecs.open(file_name, "w", encoding, "strict", -1)

            # Read everything
            # CAUTION : 2.7 return None :(
            return rd.write(text_buffer)
        except IOError as e:
            # Exception...
            logger.warning("append_text_to_file : IOError, ex=%s", SolBase.extostr(e))
            return -1
        except Exception as e:
            logger.warning("append_text_to_file : Exception, ex=%s", SolBase.extostr(e))
            return -1
        finally:
            # Close if not None...
            if rd:
                rd.close()



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

from logging.handlers import SysLogHandler
import socket

from gevent import GreenletExit

from pysolbase.SolBase import SolBase


# noinspection PyPep8
class SysLogger(SysLogHandler):
    """
    Sys log handler (will format and emit logs toward rsyslog)
    """

    def __init__(self, address="/dev/log", facility=SysLogHandler.LOG_LOCAL1, socktype=socket.SOCK_DGRAM,
                 log_callback=None):
        """
        Init
        :param address: tuple ('ip', port) or string "target"
        :type address:  str, tuple
        :param facility: log facility
        :type facility: int
        :param socktype: Type of socket
        :type socktype: socket.SocketKind
        :param log_callback: Callback for unit test
        """

        # To avoid some warnings
        self.socket = None
        self.address = address

        # Store
        self._log_callback = log_callback
        self.socktype = socktype

        # Base call
        SysLogHandler.__init__(self, address, facility)

    def notify_log(self, msg):
        """
        Notify log to callback if set (unittest purpose)
        :param msg: Log message
        :type msg: str,bytes
        """
        # noinspection PyBroadException
        try:
            if self._log_callback:
                self._log_callback(msg)
        except:
            pass

    def emit(self, record):
        """
        Emit a record.
        :param record: The record to log
        :type record: logging.LogRecord
        """

        # Write
        # noinspection PyBroadException
        try:
            # Format
            msg = self.format(record) + '\000'

            # Get component name
            cn = SolBase.get_compo_name()

            # We append the machine+component name at the beginning
            msg = u"{0} | {1} | {2}".format(SolBase.get_machine_name(), cn, msg)
            msg = msg.encode('utf-8')

            # Add priority + facility (int)
            priority = '<%d>' % self.encodePriority(self.facility, self.mapPriority(record.levelname))
            priority = priority.encode('utf-8')

            # Cn
            cn = cn.encode('utf-8')

            # noinspection PyAugmentAssignment
            msg = priority + cn + b": " + msg

            # Notify
            self.notify_log(msg)

            # Send to socket
            if self.unixsocket:
                try:
                    self.socket.send(msg)
                except socket.error:
                    # noinspection PyUnresolvedReferences
                    self._connect_unixsocket(self.address)
                    self.socket.send(msg)
            elif self.socktype == socket.SOCK_DGRAM:
                self.socket.sendto(msg, self.address)
            else:
                self.socket.sendall(msg)
        except GreenletExit:
            pass
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)
